# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
import os
import inspect
import threading
import time

from collections import namedtuple
from asyncio import sleep as asleep

import pytest

from tango import DeviceProxy, GreenMode, DevFailed
from tango.log4tango import (
    LogIt,
    DebugIt,
    InfoIt,
    WarnIt,
    ErrorIt,
    FatalIt,
    TangoStream,
)
from tango.test_utils import general_asyncio_decorator, general_decorator
from tango.server import Device, attribute, command
from tango.test_context import DeviceTestContext, MultiDeviceTestContext

# Constants
WINDOWS = "nt" in os.name

LogConsumerEntry = namedtuple(
    "LogConsumerEntry",
    ["timestamp", "level", "source", "message", "unused", "thread_id"],
)


def check_log(
    received_logs, device_name, log_level, method_name, arg, kwarg, ret, error
):

    entry_log = LogConsumerEntry(*received_logs[-2])
    exit_log = LogConsumerEntry(*received_logs[-1])

    assert entry_log.source == "test/logdevice/1"
    assert exit_log.source == "test/logdevice/1"
    assert log_level == entry_log.level
    assert log_level == exit_log.level
    if kwarg != "":
        assert f"-> {device_name}.{method_name}({arg}, {kwarg})" == entry_log.message
    else:
        assert f"-> {device_name}.{method_name}({arg})" == entry_log.message
    if error != "":
        assert f"<- {device_name}.{method_name}() raised exception!" == exit_log.message
    elif ret != "":
        assert f"{ret} <- {device_name}.{method_name}()" == exit_log.message
    else:
        assert f"<- {device_name}.{method_name}()" == exit_log.message


class LogReceiver(Device):

    _received_logs = []

    @command
    def log(self, str_in: list[str]):
        self._received_logs.append(str_in)

    @command
    def log_size(self) -> int:
        return len(self._received_logs)

    @command
    def check_last_log(self, data_to_compare: list[str,]):
        check_log(self._received_logs, *data_to_compare)
        self._received_logs = []


class AsyncLogReceiver(Device):
    green_mode = GreenMode.Asyncio

    _received_logs = []

    @command
    async def log(self, str_in: list[str]):
        self._received_logs.append(str_in)

    @command
    async def log_size(self) -> int:
        return len(self._received_logs)

    @command
    async def check_last_log(self, data_to_compare: list[str,]):
        check_log(self._received_logs, *data_to_compare)
        self._received_logs = []


# if we do not pause in method, sometimes method out log can come earlier, than method in
method_pause = 0.01


class LogDevice(Device):

    @command
    @LogIt()
    def log_cmd(self):
        time.sleep(method_pause)

    @attribute
    @DebugIt()
    def debug_attr(self) -> int:
        time.sleep(method_pause)
        return 1

    @attribute
    @InfoIt()
    def info_attr(self) -> int:
        time.sleep(method_pause)
        return 1

    @attribute
    @WarnIt()
    def warn_attr(self) -> int:
        time.sleep(method_pause)
        return 1

    @attribute
    @ErrorIt()
    def error_attr(self) -> int:
        time.sleep(method_pause)
        return 1

    @attribute
    @FatalIt()
    def fatal_attr(self) -> int:
        time.sleep(method_pause)
        return 1

    @LogIt(show_kwargs=True, show_args=True, show_ret=True)
    def arg_method(self, arg: str, kwarg: str):
        time.sleep(method_pause)
        return arg + kwarg

    @command
    def arg_kwarg_test(self, test_long_arg: bool):
        if test_long_arg:
            self.arg_method("a" * 26, kwarg="b" * 26)
        else:
            self.arg_method("a", kwarg="b")

    @command
    @LogIt()
    def error_test(self):
        time.sleep(method_pause)
        raise RuntimeError("Test error!")


class AsyncLogDevice(Device):
    green_mode = GreenMode.Asyncio

    @command
    @LogIt()
    async def log_cmd(self):
        await asleep(method_pause)

    @LogIt(show_kwargs=True, show_args=True, show_ret=True)
    async def arg_method(self, arg: int, kwarg: int):
        await asleep(method_pause)
        return arg + kwarg

    @command
    async def arg_kwarg_test(self):
        await self.arg_method(1, kwarg=2)

    @command
    @LogIt()
    async def error_test(self):
        await asleep(method_pause)
        raise RuntimeError("Test error!")


delay = 0.1
max_attempts = 10


def check_logs(log_device, expected_log):
    attempt = 0
    while log_device.log_size() < 2 and attempt < max_attempts:
        attempt += 1
        time.sleep(delay)

    if log_device.log_size() < 2:
        raise RuntimeError(
            f"Cannot receive logs for {expected_log[0]}:{expected_log[2]}"
        )

    log_device.check_last_log(expected_log)


devices_info = (
    {
        "class": LogReceiver,
        "devices": [
            {"name": "test/logreceiver/1"},
        ],
    },
    {
        "class": LogDevice,
        "devices": [
            {"name": "test/logdevice/1"},
        ],
    },
)


def test_logging_decorators():
    with MultiDeviceTestContext(devices_info) as context:
        test_device = DeviceProxy("test/logdevice/1")
        log_device = DeviceProxy("test/logreceiver/1")

        test_device.add_logging_target(
            f"device::{context.get_device_access('test/logreceiver/1')}"
        )
        test_device.log_cmd()
        check_logs(log_device, ("LogDevice", "DEBUG", "log_cmd", "", "", "", ""))

        for log_level, attr in (
            ("DEBUG", "debug_attr"),
            ("WARN", "warn_attr"),
            ("ERROR", "error_attr"),
            ("FATAL", "fatal_attr"),
            ("INFO", "info_attr"),
        ):
            getattr(test_device, attr)
            check_logs(log_device, ("LogDevice", log_level, attr, "", "", "", ""))

        test_device.arg_kwarg_test(False)
        check_logs(
            log_device,
            ("LogDevice", "DEBUG", "arg_method", "'a'", "kwarg='b'", "'ab'", ""),
        )

        test_device.arg_kwarg_test(True)
        check_logs(
            log_device,
            (
                "LogDevice",
                "DEBUG",
                "arg_method",
                "'" + "a" * 19 + "[...]",
                "kwarg='" + "b" * 19 + "[...]",
                "'" + "a" * 19 + "[...]",
                "",
            ),
        )

        with pytest.raises(DevFailed):
            test_device.error_test()
        check_logs(
            log_device, ("LogDevice", "DEBUG", "error_test", "", "", "", "Test error!")
        )


async_devices_info = (
    {
        "class": AsyncLogReceiver,
        "devices": [
            {"name": "test/logreceiver/1"},
        ],
    },
    {
        "class": AsyncLogDevice,
        "devices": [
            {"name": "test/logdevice/1"},
        ],
    },
)


def test_async_logging_decorators():
    with MultiDeviceTestContext(async_devices_info) as context:
        log_emitter = DeviceProxy("test/logdevice/1")
        log_receiver = DeviceProxy("test/logreceiver/1")

        log_emitter.add_logging_target(
            f"device::{context.get_device_access('test/logreceiver/1')}"
        )
        log_emitter.log_cmd()
        check_logs(log_receiver, ("AsyncLogDevice", "DEBUG", "log_cmd", "", "", "", ""))

        log_emitter.arg_kwarg_test()
        check_logs(
            log_receiver,
            ("AsyncLogDevice", "DEBUG", "arg_method", "1", "kwarg=2", "3", ""),
        )

        with pytest.raises(DevFailed):
            log_emitter.error_test()
        check_logs(
            log_receiver,
            ("AsyncLogDevice", "DEBUG", "error_test", "", "", "", "Test error!"),
        )


class Receiver:
    n_calls = 0

    def __call__(self, stream):
        self.n_calls += 1
        assert stream == "ab\nc"

    def check_calls(self, expected_n):
        assert self.n_calls == expected_n


def test_tango_stream():
    receiver = Receiver()
    stream = TangoStream(receiver)
    stream.write("ab")
    stream.write("\nc\n")

    receiver.check_calls(1)

    stream.flush()
    receiver.check_calls(1)

    stream.write("ab\nc")
    receiver.check_calls(2)


def test_logging(server_green_mode):
    log_received = threading.Event()

    if server_green_mode == GreenMode.Asyncio:

        class LogSourceDevice(Device):
            green_mode = server_green_mode
            _last_log_time = 0.0

            @command(dtype_in=("str",))
            async def log_fatal_message(self, msg):
                self._last_log_time = time.time()
                if len(msg) > 1:
                    self.fatal_stream(msg[0], msg[1])
                else:
                    self.fatal_stream(msg[0])

            @command(dtype_in=("str",))
            async def log_error_message(self, msg):
                self._last_log_time = time.time()
                if len(msg) > 1:
                    self.error_stream(msg[0], msg[1])
                else:
                    self.error_stream(msg[0])

            @command(dtype_in=("str",))
            async def log_warn_message(self, msg):
                self._last_log_time = time.time()
                if len(msg) > 1:
                    self.warn_stream(msg[0], msg[1])
                else:
                    self.warn_stream(msg[0])

            @command(dtype_in=("str",))
            async def log_info_message(self, msg):
                self._last_log_time = time.time()
                if len(msg) > 1:
                    self.info_stream(msg[0], msg[1])
                else:
                    self.info_stream(msg[0])

            @command(dtype_in=("str",))
            async def log_debug_message(self, msg):
                self._last_log_time = time.time()
                if len(msg) > 1:
                    self.debug_stream(msg[0], msg[1])
                else:
                    self.debug_stream(msg[0])

            @attribute(dtype=float)
            async def last_log_time(self):
                return self._last_log_time

        class LogConsumerDevice(Device):
            green_mode = server_green_mode
            _last_log_data = []

            @command(dtype_in=("str",))
            async def Log(self, argin):
                self._last_log_data = argin
                log_received.set()

            @attribute(dtype=int)
            async def last_log_timestamp_ms(self):
                return int(self._last_log_data[0])

            @attribute(dtype=str)
            async def last_log_level(self):
                return self._last_log_data[1]

            @attribute(dtype=str)
            async def last_log_source(self):
                return self._last_log_data[2]

            @attribute(dtype=str)
            async def last_log_message(self):
                return self._last_log_data[3]

            @attribute(dtype=str)
            async def last_log_context_unused(self):
                return self._last_log_data[4]

            @attribute(dtype=str)
            async def last_log_thread_id(self):
                return self._last_log_data[5]

    else:

        class LogSourceDevice(Device):
            green_mode = server_green_mode
            _last_log_time = 0.0

            @command(dtype_in=("str",))
            def log_fatal_message(self, msg):
                self._last_log_time = time.time()
                if len(msg) > 1:
                    self.fatal_stream(msg[0], msg[1])
                else:
                    self.fatal_stream(msg[0])

            @command(dtype_in=("str",))
            def log_error_message(self, msg):
                self._last_log_time = time.time()
                if len(msg) > 1:
                    self.error_stream(msg[0], msg[1])
                else:
                    self.error_stream(msg[0])

            @command(dtype_in=("str",))
            def log_warn_message(self, msg):
                self._last_log_time = time.time()
                if len(msg) > 1:
                    self.warn_stream(msg[0], msg[1])
                else:
                    self.warn_stream(msg[0])

            @command(dtype_in=("str",))
            def log_info_message(self, msg):
                self._last_log_time = time.time()
                if len(msg) > 1:
                    self.info_stream(msg[0], msg[1])
                else:
                    self.info_stream(msg[0])

            @command(dtype_in=("str",))
            def log_debug_message(self, msg):
                self._last_log_time = time.time()
                if len(msg) > 1:
                    self.debug_stream(msg[0], msg[1])
                else:
                    self.debug_stream(msg[0])

            @attribute(dtype=float)
            def last_log_time(self):
                return self._last_log_time

        class LogConsumerDevice(Device):
            green_mode = server_green_mode
            _last_log_data = []

            @command(dtype_in=("str",))
            def Log(self, argin):
                self._last_log_data = argin
                log_received.set()

            @attribute(dtype=int)
            def last_log_timestamp_ms(self):
                return int(self._last_log_data[0])

            @attribute(dtype=str)
            def last_log_level(self):
                return self._last_log_data[1]

            @attribute(dtype=str)
            def last_log_source(self):
                return self._last_log_data[2]

            @attribute(dtype=str)
            def last_log_message(self):
                return self._last_log_data[3]

            @attribute(dtype=str)
            def last_log_context_unused(self):
                return self._last_log_data[4]

            @attribute(dtype=str)
            def last_log_thread_id(self):
                return self._last_log_data[5]

    def assert_log_details_correct(level, msg):
        assert log_received.wait(0.5)
        _assert_log_time_close_enough()
        _assert_log_fields_correct_for_level(level, msg)
        log_received.clear()

    def _assert_log_time_close_enough():
        log_emit_time = proxy_source.last_log_time
        log_receive_time = proxy_consumer.last_log_timestamp_ms / 1000.0
        now = time.time()
        # cppTango logger time function may use a different
        # implementation to CPython's time.time().  This is
        # especially noticeable on Windows platforms.
        timer_implementation_tolerance = 0.020 if WINDOWS else 0.001
        min_time = log_emit_time - timer_implementation_tolerance
        max_time = now + timer_implementation_tolerance
        assert min_time <= log_receive_time <= max_time

    def _assert_log_fields_correct_for_level(level, msg):
        assert proxy_consumer.last_log_level == level.upper()
        assert proxy_consumer.last_log_source == "test/log/source"
        assert proxy_consumer.last_log_message == msg
        assert proxy_consumer.last_log_context_unused == ""
        assert len(proxy_consumer.last_log_thread_id) > 0

    devices_info = (
        {"class": LogSourceDevice, "devices": [{"name": "test/log/source"}]},
        {"class": LogConsumerDevice, "devices": [{"name": "test/log/consumer"}]},
    )

    with MultiDeviceTestContext(devices_info) as context:
        proxy_source = context.get_device("test/log/source")
        proxy_consumer = context.get_device("test/log/consumer")
        consumer_access = context.get_device_access("test/log/consumer")
        proxy_source.add_logging_target(f"device::{consumer_access}")

        for msg in ([""], [" with literal %s"], [" with string %s as arg", "foo"]):
            level = "fatal"
            log_msg = msg[:]
            log_msg[0] = "test " + level + msg[0]
            proxy_source.log_fatal_message(log_msg)
            if len(msg) > 1:
                check_log_msg = log_msg[0] % log_msg[1]
            else:
                check_log_msg = log_msg[0]
            assert_log_details_correct(level, check_log_msg)

            level = "error"
            log_msg = msg[:]
            log_msg[0] = "test " + level + msg[0]
            proxy_source.log_error_message(log_msg)
            if len(msg) > 1:
                check_log_msg = log_msg[0] % log_msg[1]
            else:
                check_log_msg = log_msg[0]
            assert_log_details_correct(level, check_log_msg)

            level = "warn"
            log_msg = msg[:]
            log_msg[0] = "test " + level + msg[0]
            proxy_source.log_warn_message(log_msg)
            if len(msg) > 1:
                check_log_msg = log_msg[0] % log_msg[1]
            else:
                check_log_msg = log_msg[0]
            assert_log_details_correct(level, check_log_msg)

            level = "info"
            log_msg = msg[:]
            log_msg[0] = "test " + level + msg[0]
            proxy_source.log_info_message(log_msg)
            if len(msg) > 1:
                check_log_msg = log_msg[0] % log_msg[1]
            else:
                check_log_msg = log_msg[0]
            assert_log_details_correct(level, check_log_msg)

            level = "debug"
            log_msg = msg[:]
            log_msg[0] = "test " + level + msg[0]
            proxy_source.log_debug_message(log_msg)
            if len(msg) > 1:
                check_log_msg = log_msg[0] % log_msg[1]
            else:
                check_log_msg = log_msg[0]
            assert_log_details_correct(level, check_log_msg)


@pytest.mark.skipif(
    not os.environ.get("PYTHONUNBUFFERED"),
    reason="This test requires PYTHONUNBUFFERED=1 to capture the outputs.",
)
def test_decorator_logging_source_location(server_green_mode, capfd):
    """Run decorated commands and attributes and verify that @InfoIt decorator
    always logs the correct location."""

    if server_green_mode == GreenMode.Asyncio:

        class InfoItDevice(Device):
            green_mode = server_green_mode

            @command(dtype_out=int)
            @InfoIt()
            async def decorated_command(self):
                return inspect.currentframe().f_lineno

            @command(dtype_out=int)
            async def run_decorated_method(self):
                return await self.decorated_method()

            @InfoIt()
            @general_asyncio_decorator
            async def decorated_method(self):
                return inspect.currentframe().f_lineno

            @attribute(dtype=int)
            @InfoIt()
            async def decorated_attribute(self):
                return inspect.currentframe().f_lineno

    else:

        class InfoItDevice(Device):
            green_mode = server_green_mode

            @command(dtype_out=int)
            @InfoIt()
            def decorated_command(self):
                return inspect.currentframe().f_lineno

            @command(dtype_out=int)
            def run_decorated_method(self):
                return self.decorated_method()

            @InfoIt()
            @general_decorator
            def decorated_method(self):
                return inspect.currentframe().f_lineno

            @attribute(dtype=int)
            @InfoIt()
            def decorated_attribute(self):
                return inspect.currentframe().f_lineno

    with DeviceTestContext(InfoItDevice, debug=3) as device:
        filename = os.path.basename(__file__)
        for cmd, method in [
            ("decorated_command", "decorated_command"),
            ("run_decorated_method", "decorated_method"),
        ]:
            lineno = device.command_inout(cmd) - 3
            out, err = capfd.readouterr()  # calling this function clears the buffer
            assert (
                f"({filename}:{lineno}) test/nodb/infoitdevice -> InfoItDevice.{method}()"
                in out
            )

        lineno = device.decorated_attribute - 3
        out, err = capfd.readouterr()
        assert (
            f"({filename}:{lineno}) test/nodb/infoitdevice -> InfoItDevice.decorated_attribute()"
            in out
        )


@pytest.mark.skipif(
    not os.environ.get("PYTHONUNBUFFERED"),
    reason="This test requires PYTHONUNBUFFERED=1 to capture the outputs.",
)
def test_stream_logging_source_location(server_green_mode, capfd):
    if server_green_mode == GreenMode.Asyncio:

        class StreamLogsDevice(Device):
            green_mode = server_green_mode

            @command(dtype_out=int)
            async def log_streams(self):
                self.info_stream("info")
                self.debug_stream("debug")
                self.warn_stream("warn")
                self.error_stream("error")
                self.fatal_stream("fatal")
                return inspect.currentframe().f_lineno

    else:

        class StreamLogsDevice(Device):
            green_mode = server_green_mode

            @command(dtype_out=int)
            def log_streams(self):
                self.info_stream("info")
                self.debug_stream("debug")
                self.warn_stream("warn")
                self.error_stream("error")
                self.fatal_stream("fatal")
                return inspect.currentframe().f_lineno

    with DeviceTestContext(StreamLogsDevice, debug=3) as device:
        lineno = device.command_inout("log_streams") - 5
        filename = os.path.basename(__file__)
        out, err = capfd.readouterr()
        for i, level in [
            (0, "INFO"),
            (1, "DEBUG"),
            (2, "WARN"),
            (3, "ERROR"),
            (4, "FATAL"),
        ]:
            assert (
                f"{level} ({filename}:{lineno + i}) test/nodb/streamlogsdevice" in out
            )



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
"""
This is an internal PyTango module.
"""

__all__ = ("group_reply_init",)

__docformat__ = "restructuredtext"

from tango.pytango_pprint import __indented, __str_error_stack_helper
from tango import GroupReply, GroupCmdReply, GroupAttrReply


def __GroupCmdReply__get_data(self):
    """
    get_data(self) -> any

        Get the actual value stored in the GroupCmdRply, the command
        output value.
        It's the same as self.get_data_raw().extract()

    Parameters : None
    Return     : (any) Whatever is stored there, or None.
    """
    return self.get_data_raw().extract()


def __str_group_reply_helper(self):
    dev_name = self.dev_name()
    obj_name = self.obj_name()
    enabled = self.group_element_enabled()
    has_failed = self.has_failed()
    if not enabled:
        extra_line = ""
    elif has_failed:
        err_str = __str_error_stack_helper(self.get_err_stack())
        err_str = f"[\n{err_str}\n]"
        extra_line = f"err_stack = {err_str}\n"
    elif hasattr(self, "get_data"):
        value = self.get_data()
        if isinstance(value, str):
            value = f'"{value}"'
        else:
            value = str(value)
        extra_line = f"data = {value}\n"
    else:
        extra_line = ""
    details = (
        f'dev_name = "{dev_name}"\n'
        f'obj_name = "{obj_name}"\n'
        f"enabled = {enabled}\n"
        f"has_failed = {has_failed}\n"
        f"{extra_line}"
    )
    return (
        f"{self.__class__.__name__}[\n"
        f"{__indented(details, strip_outer=False)}\n"
        f"]"
    )


def group_reply_init():
    GroupCmdReply.get_data = __GroupCmdReply__get_data
    GroupCmdReply.__str__ = __str_group_reply_helper
    GroupAttrReply.__str__ = __str_group_reply_helper
    GroupReply.__str__ = __str_group_reply_helper



#!/usr/bin/env python3
# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

"""
Trivial power supply device with no external connection or behaviour.

Extends ps0a.py, adding noise to the voltage attribute and configuring
it to be polled in the server code.
"""

import random
from time import sleep
from tango.server import Device, attribute, command


class PowerSupply(Device):
    @attribute(dtype=float, polling_period=3000, rel_change=1e-3)  # milliseconds
    def voltage(self):
        noise = -0.05 + 0.1 * random.random()
        return 1.5 + noise

    @command
    def calibrate(self):
        sleep(0.1)


if __name__ == "__main__":
    PowerSupply.run_server()



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
"""
A module defining pytest fixtures for testing with MultiDeviceTestContext
Requires pytest, and at least PyTango 9.5.0
(see commit history for the approach to with earlier versions)
"""

from collections import defaultdict
import pytest
from tango.test_context import MultiDeviceTestContext


@pytest.fixture(scope="module")
def devices_info(request):
    yield getattr(request.module, "devices_info")


@pytest.fixture(scope="function")
def tango_context(devices_info):
    """
    Creates and returns a TANGO MultiDeviceTestContext object.
    """
    with MultiDeviceTestContext(devices_info, process=True) as context:
        yield context



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
# -*- coding: utf-8 -*-

import pytest

from typing import List

import operator

from tango import (
    StdStringVector,
    StdDoubleVector,
    DbDevInfos,
    DbDevInfo,
    DbDevImportInfos,
    DbDevImportInfo,
    DbDevExportInfo,
    DbDevExportInfos,
    DbHistory,
    DbHistoryList,
    DbData,
    DbDatum,
    DeviceData,
    DeviceDataList,
    DeviceDataHistory,
    DeviceDataHistoryList,
)

from tango.utils import (
    _clear_test_context_tango_host_fqtrl,
    _get_device_fqtrl_if_necessary,
    _get_test_context_tango_host_fqtrl,
    _set_test_context_tango_host_fqtrl,
    _traced_coverage_run_active,
    InvalidTangoHostTrlError,
    StdStringVector_2_seq,
    StdDoubleVector_2_seq,
    seq_2_StdStringVector,
    seq_2_StdDoubleVector,
    seq_2_DbDevInfos,
    seq_2_DbDevExportInfos,
    seq_2_DbData,
    get_tango_type,
    TO_TANGO_TYPE,
)


@pytest.fixture()
def restore_global():
    yield
    _clear_test_context_tango_host_fqtrl()


@pytest.mark.parametrize(
    "override_trl, input_trl, expected_trl",
    [
        (None, "a/b/c", "a/b/c"),
        (None, "a/b/c/d", "a/b/c/d"),
        (None, "tango://host:12/a/b/c", "tango://host:12/a/b/c"),
        (None, "tango://host:12/a/b/c#dbase=no", "tango://host:12/a/b/c#dbase=no"),
        (None, "no://trl/validation", "no://trl/validation"),
        ("tango://host:12", "a/b/c", "tango://host:12/a/b/c"),
        ("tango://host:12", "a/b/c/d", "tango://host:12/a/b/c/d"),
        ("tango://host:12", "tango://host:12/a/b/c", "tango://host:12/a/b/c"),
        ("tango://host:12#dbase=no", "a/b/c", "tango://host:12/a/b/c#dbase=no"),
        ("tango://host:12#dbase=yes", "a/b/c", "tango://host:12/a/b/c#dbase=yes"),
        ("tango://127.0.0.1:12", "a/b/c", "tango://127.0.0.1:12/a/b/c"),
    ],
)
def test_get_trl_with_test_fqtrl_success(
    override_trl, input_trl, expected_trl, restore_global
):
    _set_test_context_tango_host_fqtrl(override_trl)
    actual_trl = _get_device_fqtrl_if_necessary(input_trl)
    assert actual_trl == expected_trl


@pytest.mark.parametrize(
    "override_trl, input_trl",
    [
        ("host:123", "a/b/c"),  # missing scheme
        ("tango://", "a/b/c"),  # missing hostname and port
        ("tango://:123", "a/b/c"),  # missing hostname
        ("tango://host", "a/b/c"),  # missing port
        ("tango://host:0", "a/b/c"),  # zero-value port
        ("tango://host:12/path", "a/b/c"),  # non-empty path
        ("tango://host:123?query=1", "a/b/c"),  # non-empty query
        ("tango://host:123#dbase=invalid", "a/b/c"),  # invalid fragment
    ],
)
def test_get_trl_with_test_fdqn_failure(override_trl, input_trl, restore_global):
    _set_test_context_tango_host_fqtrl(override_trl)
    with pytest.raises(InvalidTangoHostTrlError):
        _ = _get_device_fqtrl_if_necessary(input_trl)


def test_global_state_default_set_and_clear(restore_global):
    default = _get_test_context_tango_host_fqtrl()
    _set_test_context_tango_host_fqtrl("tango://localhost:1234")
    after_set = _get_test_context_tango_host_fqtrl()
    _clear_test_context_tango_host_fqtrl()
    after_clear = _get_test_context_tango_host_fqtrl()

    assert default is None
    assert after_set == "tango://localhost:1234"
    assert after_clear is None


def test_clear_global_var_without_set_does_not_raise():
    _clear_test_context_tango_host_fqtrl()


def test_get_tango_type_valid():
    from tango import DevString, DevLong64, AttrDataFormat

    assert get_tango_type("abc") == (DevString, AttrDataFormat.SCALAR)
    assert get_tango_type(123) == (DevLong64, AttrDataFormat.SCALAR)
    assert get_tango_type([1, 2, 3]) == (DevLong64, AttrDataFormat.SPECTRUM)
    assert get_tango_type([[1, 2, 3], [4, 5, 6]]) == (DevLong64, AttrDataFormat.IMAGE)


def test_get_tango_type_invalid_raises_type_error():
    class NonTangoType:
        pass

    with pytest.raises(TypeError):
        get_tango_type(NonTangoType())
    with pytest.raises(TypeError):
        get_tango_type([{"start with": "invalid type"}, "abc", 123])
    # TODO: check data type for all nested items.  E.g., this doesn't raise TypeError:  ["abc", 123, {"k": "v"}]


def check_vector_to_seq_conversion(
    vec_type, seq_type, vec_2_seq_func, seq_2_vec_func, elem, equal_op=operator.eq
):

    vec = vec_type()

    ret = seq_2_vec_func(vec)
    assert vec is ret

    ret = seq_2_vec_func([elem], vec)
    assert vec is ret
    assert isinstance(ret, vec_type)
    assert len(ret) == 1
    assert equal_op(ret[0], elem)

    ret = seq_2_vec_func([elem])
    assert isinstance(ret, vec_type)
    assert len(ret) == 1
    assert equal_op(ret[0], elem)

    with pytest.raises(TypeError, match="vec must be"):
        seq_2_vec_func([], [])

    if vec_2_seq_func is not None:

        vec = vec_type()
        vec.append(elem)

        ret = vec_2_seq_func(vec)
        assert isinstance(ret, List)
        assert len(ret) == 1
        assert equal_op(ret[0], elem)

        seq = []
        ret = vec_2_seq_func(vec, seq=seq)
        assert seq is ret
        assert isinstance(ret, List)
        assert len(ret) == 1
        assert equal_op(ret[0], elem)

        with pytest.raises(TypeError, match="vec must be"):
            vec_2_seq_func([], [])


def test_sequence_to_string_vector_and_back():

    check_vector_to_seq_conversion(
        StdStringVector, List, StdStringVector_2_seq, seq_2_StdStringVector, "abcd"
    )


def test_sequence_to_double_vector_and_back():

    check_vector_to_seq_conversion(
        StdDoubleVector, List, StdDoubleVector_2_seq, seq_2_StdDoubleVector, 123.0
    )


def test_sequence_to_dbdevinfo_vector_and_back():

    info = DbDevInfo()
    info.klass = "a"
    info.name = "b"
    info.server = "c"

    def equal_op(left, right):
        return (
            left.klass == right.klass
            and left.name == right.name
            and left.server == right.server
        )

    check_vector_to_seq_conversion(
        DbDevInfos, List, None, seq_2_DbDevInfos, info, equal_op=equal_op
    )


def test_sequence_to_dbdevexportinfo_vector_and_back():

    info = DbDevExportInfo()
    info.name = "a"
    info.ior = "b"
    info.host = "c"
    info.version = "d"
    info.pid = 123

    def equal_op(left, right):
        return (
            left.name == right.name
            and left.ior == right.ior
            and left.host == right.host
            and left.version == right.version
            and left.pid == right.pid
        )

    check_vector_to_seq_conversion(
        DbDevExportInfos, List, None, seq_2_DbDevExportInfos, info, equal_op=equal_op
    )


def test_sequence_to_dbdata_vector_and_back():

    info = DbDatum()
    info.name = "a"

    def equal_op(left, right):
        return left.name == right.name

    check_vector_to_seq_conversion(
        DbData, List, None, seq_2_DbData, info, equal_op=equal_op
    )


def seq_2_x(vec_type, seq, vec=None):
    if vec is None:
        if isinstance(seq, vec_type):
            return seq
        vec = vec_type()
    if not isinstance(vec, vec_type):
        raise TypeError("vec must be a ...")
    for e in seq:
        vec.append(e)
    return vec


def seq_2_DbDevImportInfos(seq, vec=None):
    return seq_2_x(DbDevImportInfos, seq, vec=vec)


def test_sequence_to_dbdevimport_vector_and_back():

    info = DbDevImportInfo()

    def equal_op(left, right):
        return (
            left.name == right.name
            and left.exported == right.exported
            and left.ior == right.ior
            and left.version == right.version
        )

    check_vector_to_seq_conversion(
        DbDevImportInfos, List, None, seq_2_DbDevImportInfos, info, equal_op=equal_op
    )


def seq_2_DbHistoryList(seq, vec=None):
    return seq_2_x(DbHistoryList, seq, vec=vec)


def test_sequence_to_dbhistorylist_vector_and_back():

    # second string, the date, is long enough that DbHistory::format_mysql_date in cppTango does not barf
    hist = DbHistory("a", "b" * 12, StdStringVector())

    def equal_op(left, right):
        return (
            left.get_attribute_name() == right.get_attribute_name()
            and left.get_date() == right.get_date()
            and left.get_name() == right.get_name()
            and left.get_value().name == right.get_value().name
        )

    check_vector_to_seq_conversion(
        DbHistoryList, List, None, seq_2_DbHistoryList, hist, equal_op=equal_op
    )


def seq_2_DeviceDataList(seq, vec=None):
    return seq_2_x(DeviceDataList, seq, vec=vec)


def test_sequence_to_devicedatalist_vector_and_back():

    dd = DeviceData()
    dd.insert(TO_TANGO_TYPE[int], 1)

    def equal_op(left, right):
        return left.is_empty() == right.is_empty()

    check_vector_to_seq_conversion(
        DeviceDataList, List, None, seq_2_DeviceDataList, dd, equal_op=equal_op
    )


def seq_2_DeviceDataHistoryList(seq, vec=None):
    return seq_2_x(DeviceDataHistoryList, seq, vec=vec)


def test_sequence_to_devicedatahistorylist_vector_and_back():

    ddh = DeviceDataHistory()

    def equal_op(left, right):
        return left.has_failed() == right.has_failed()

    check_vector_to_seq_conversion(
        DeviceDataHistoryList,
        List,
        None,
        seq_2_DeviceDataHistoryList,
        ddh,
        equal_op=equal_op,
    )


def test_report_coverage_tracing_enabled():
    # The point of this test is to have something in the test
    # output that shows if coverage tracing is active or not.
    # If it isn't active, we skip the test.
    if not _traced_coverage_run_active:
        pytest.skip("Coverage tracing is disabled")
    assert _traced_coverage_run_active



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

"""
This is an internal PyTango module.
"""

__all__ = ("base_types_init",)

__docformat__ = "restructuredtext"

from tango import (
    StdStringVector,
    StdLongVector,
    StdDoubleVector,
    CommandInfoList,
    AttributeInfoList,
    AttributeInfoListEx,
    DbData,
    DbDevInfos,
    DbDevExportInfos,
    DbDevImportInfos,
    DbHistoryList,
    DeviceDataHistoryList,
)


def __StdVector__add(self, seq):
    ret = seq.__class__(self)
    ret.extend(seq)
    return ret


def __StdVector__mul(self, n):
    ret = self.__class__()
    for _ in range(n):
        ret.extend(self)
    return ret


def __StdVector__imul(self, n):
    ret = self.__class__()
    for _ in range(n):
        ret.extend(self)
    return ret


def __fillVectorClass(klass):
    klass.__add__ = __StdVector__add
    klass.__mul__ = __StdVector__mul
    klass.__imul__ = __StdVector__imul


def base_types_init():
    v_klasses = (
        StdStringVector,
        StdLongVector,
        StdDoubleVector,
        CommandInfoList,
        AttributeInfoList,
        AttributeInfoListEx,
        DbData,
        DbDevInfos,
        DbDevExportInfos,
        DbDevImportInfos,
        DbHistoryList,
        DeviceDataHistoryList,
    )

    for v_klass in v_klasses:
        __fillVectorClass(v_klass)

    # Doc string for vectors is easier to add already in Python

    AttributeInfoList.__doc__ = """
    List of AttributeInfo objects, containing available information for the attributes"""

    AttributeInfoListEx.__doc__ = """
    List of AttributeInfoEx objects, containing available information for the attributes"""



#!/usr/bin/env python3
# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
#
# client example: reading an attribute, also with simplified (high-level) syntax

import sys
from tango import DeviceProxy, DevFailed, Except

dev_name = "sys/tg_test/1"
attr_name = "double_scalar"

# step 1 - create the device proxy
try:
    dev = DeviceProxy(dev_name)
    print(f"proxy for {dev_name} created")
except DevFailed as ex:
    Except.print_exception(ex)
    sys.exit(1)

# step 2 - read the attribute
try:
    da = dev.read_attribute(attr_name)  # read the attribute
    value = da.value
except DevFailed as ex:
    Except.print_exception(ex)
    print("failed to read attribute")
    sys.exit(1)

print(f"{dev_name}/{attr_name} value: {value}")

# step 3 - use simplified interface
try:
    value2 = dev.double_scalar
except DevFailed as ex:
    Except.print_exception(ex)
    print("failed to read double_scalar")
    sys.exit(1)

print(f"{dev_name}/{attr_name} value2: {value2}")

"""
Typical output:

➜  training git:(develop) ✗ docker-compose exec cli python /training/client/reading01.py
proxy for sys/tg_test/1 created
sys/tg_test/1/double_scalar value: -184.15039639576005
sys/tg_test/1/double_scalar value2: -187.22596334769938
"""



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

"""
This is an internal PyTango module.
"""


__all__ = ("DeviceClass", "device_class_init")

__docformat__ = "restructuredtext"

import collections.abc

from tango._tango import (
    Except,
    DevFailed,
    DeviceClass,
    CmdArgType,
    DispLevel,
    UserDefaultAttrProp,
)
from tango.pyutil import Util

from tango.utils import is_pure_str, is_non_str_seq, seqStr_2_obj, obj_2_str, is_array

from tango.globals import get_class, get_class_by_class, get_constructed_class_by_class
from tango.attr_data import AttrData


class PropUtil:
    """An internal Property util class"""

    scalar_int_types = (
        CmdArgType.DevShort,
        CmdArgType.DevUShort,
        CmdArgType.DevLong,
        CmdArgType.DevULong,
    )

    scalar_long_types = (CmdArgType.DevLong64, CmdArgType.DevULong64)

    scalar_float_types = (
        CmdArgType.DevFloat,
        CmdArgType.DevDouble,
    )

    scalar_numerical_types = scalar_int_types + scalar_long_types + scalar_float_types

    scalar_str_types = (
        CmdArgType.DevString,
        CmdArgType.ConstDevString,
    )

    scalar_types = (
        scalar_numerical_types
        + scalar_str_types
        + (
            CmdArgType.DevBoolean,
            CmdArgType.DevEncoded,
            CmdArgType.DevUChar,
            CmdArgType.DevVoid,
        )
    )

    def __init__(self):
        self.db = None
        if Util._UseDb:
            self.db = Util.instance().get_database()

    def set_default_property_values(self, dev_class, class_prop, dev_prop):
        """
        set_default_property_values(self, dev_class, class_prop, dev_prop) -> None

            Sets the default property values

        Parameters :
            - dev_class : (DeviceClass) device class object
            - class_prop : (dict<str,>) class properties
            - dev_prop : (dict<str,>) device properties

        Return     : None
        """
        for name in class_prop:
            type = self.get_property_type(name, class_prop)
            val = self.get_property_values(name, class_prop)
            val = self.values2string(val, type)
            desc = self.get_property_description(name, class_prop)
            dev_class.add_wiz_class_prop(name, desc, val)

        for name in dev_prop:
            type = self.get_property_type(name, dev_prop)
            val = self.get_property_values(name, dev_prop)
            val = self.values2string(val, type)
            desc = self.get_property_description(name, dev_prop)
            dev_class.add_wiz_dev_prop(name, desc, val)

    def get_class_properties(self, dev_class, class_prop):
        """
        get_class_properties(self, dev_class, class_prop) -> None

                Returns the class properties

            Parameters :
                - dev_class : (DeviceClass) the DeviceClass object
                - class_prop : [in, out] (dict<str, None>) the property names. Will be filled
                               with property values

            Return     : None"""
        # initialize default values
        if class_prop == {} or not Util._UseDb:
            return

        # call database to get properties
        props = self.db.get_class_property(
            dev_class.get_name(), list(class_prop.keys())
        )

        # if value defined in database, store it
        for name in class_prop:
            if props[name]:
                type = self.get_property_type(name, class_prop)
                values = self.stringArray2values(props[name], type)
                self.set_property_values(name, class_prop, values)
            else:
                print(name + " property NOT found in database")

    def get_device_properties(self, dev, class_prop, dev_prop):
        """
        get_device_properties(self, dev, class_prop, dev_prop) -> None

                Returns the device properties

            Parameters :
                - dev : (DeviceImpl) the device object
                - class_prop : (dict<str, obj>) the class properties
                - dev_prop : [in,out] (dict<str, None>) the device property names

            Return     : None"""
        #    initialize default properties
        if dev_prop == {} or not Util._UseDb:
            return

        # Call database to get properties
        props = self.db.get_device_property(dev.get_name(), list(dev_prop.keys()))
        #    if value defined in database, store it
        for name in dev_prop:
            prop_value = props[name]
            if len(prop_value):
                data_type = self.get_property_type(name, dev_prop)
                values = self.stringArray2values(prop_value, data_type)
                if not self.is_empty_seq(values):
                    self.set_property_values(name, dev_prop, values)
                else:
                    #    Try to get it from class property
                    values = self.get_property_values(name, class_prop)
                    if not self.is_empty_seq(values):
                        if not self.is_seq(values):
                            values = [values]
                        data_type = self.get_property_type(name, class_prop)
                        values = self.stringArray2values(values, data_type)
                        if not self.is_empty_seq(values):
                            self.set_property_values(name, dev_prop, values)
            else:
                #    Try to get it from class property
                values = self.get_property_values(name, class_prop)
                if not self.is_empty_seq(values):
                    if not self.is_seq(values):
                        values = [values]
                    data_type = self.get_property_type(name, class_prop)
                    values = self.stringArray2values(values, data_type)
                    if not self.is_empty_seq(values):
                        self.set_property_values(name, dev_prop, values)

    def is_seq(self, v):
        """
        is_seq(self, v) -> bool

                Helper method. Determines if the object is a sequence

            Parameters :
                - v : (object) the object to be analysed

            Return     : (bool) True if the object is a sequence or False otherwise"""
        return isinstance(v, collections.abc.Sequence)

    def is_empty_seq(self, v):
        """
        is_empty_seq(self, v) -> bool

                Helper method. Determines if the object is an empty sequence

            Parameters :
                - v : (object) the object to be analysed

            Return     : (bool) True if the object is a sequence which is empty or False otherwise
        """
        return self.is_seq(v) and not len(v)

    def get_property_type(self, prop_name, properties):
        """
        get_property_type(self, prop_name, properties) -> CmdArgType

                Gets the property type for the given property name using the
                information given in properties

            Parameters :
                - prop_name : (str) property name
                - properties : (dict<str,data>) property data

            Return     : (CmdArgType) the tango type for the given property"""
        try:
            tg_type = properties[prop_name][0]
        except Exception:
            tg_type = CmdArgType.DevVoid
        return tg_type

    def set_property_values(self, prop_name, properties, values):
        """
        set_property_values(self, prop_name, properties, values) -> None

                Sets the property value in the properties

            Parameters :
                - prop_name : (str) property name
                - properties : (dict<str,obj>) [in,out] dict which will contain the value
                - values : (seq) the new property value

            Return     : None"""

        properties[prop_name][2] = values

    def get_property_values(self, prop_name, properties):
        """
        get_property_values(self, prop_name, properties) -> obj

                Gets the property value

            Parameters :
                - prop_name : (str) property name
                - properties : (dict<str,obj>) properties
            Return     : (obj) the value for the given property name"""
        tg_type = self.get_property_type(prop_name, properties)

        try:
            val = properties[prop_name][2]
        except Exception:
            val = []

        if is_pure_str(val):
            val = self.stringArray2values(val, tg_type)

        if is_array(tg_type) or (
            isinstance(val, collections.abc.Sequence) and not len(val)
        ):
            return val
        else:
            if is_non_str_seq(val):
                return val[0]
            else:
                return val

    def get_property_description(self, prop_name, properties):
        """
        get_property_description(self, prop_name, properties) -> obj

                Gets the property description

            Parameters :
                - prop_name : (str) property name
                - properties : (dict<str,obj>) properties
            Return     : (str) the description for the given property name"""
        return properties[prop_name][1]

    def stringArray2values(self, argin, argout_type):
        """internal helper method"""
        return seqStr_2_obj(argin, argout_type)

    def values2string(self, argin, argout_type):
        """internal helper method"""
        return obj_2_str(argin, argout_type)


def __DeviceClass__init__(self, name):
    DeviceClass.__init_orig__(self, name)
    try:
        pu = self.prop_util = PropUtil()
        self.py_dev_list = []
        pu.set_default_property_values(
            self, self.class_property_list, self.device_property_list
        )
        pu.get_class_properties(self, self.class_property_list)
        for prop_name in self.class_property_list:
            if not hasattr(self, prop_name):
                setattr(
                    self,
                    prop_name,
                    pu.get_property_values(prop_name, self.class_property_list),
                )
    except DevFailed as df:
        print(f"PyDS: {name}: A Tango error occurred in the constructor:")
        Except.print_exception(df)
    except Exception as e:
        print(f"PyDS: {name}: An error occurred in the constructor:")
        print(str(e))


def __DeviceClass__str__(self):
    return f"{self.__class__.__name__}({self.get_name()})"


def __DeviceClass__repr__(self):
    return f"{self.__class__.__name__}({self.get_name()})"


def __throw_create_attribute_exception(msg):
    """
    Helper method to throw DevFailed exception when inside
    create_attribute
    """
    Except.throw_exception("PyDs_WrongAttributeDefinition", msg, "create_attribute()")


def __throw_create_command_exception(msg):
    """
    Helper method to throw DevFailed exception when inside
    create_command
    """
    Except.throw_exception("PyDs_WrongCommandDefinition", msg, "create_command()")


def __DeviceClass__create_user_default_attr_prop(self, attr_name, extra_info):
    """for internal usage only"""
    p = UserDefaultAttrProp()
    for k, v in extra_info.items():
        k_lower = k.lower()
        method_name = f"set_{k_lower.replace(' ', '_')}"
        if hasattr(p, method_name):
            method = getattr(p, method_name)
            method(str(v))
        elif k == "delta_time":
            p.set_delta_t(str(v))
        elif k_lower not in ("display level", "polling period", "memorized"):
            name = self.get_name()
            msg = (
                f"Wrong definition of attribute {attr_name} in "
                f"class {name}\nThe object extra information '{k}' "
                f"is not recognized!"
            )
            self.__throw_create_attribute_exception(msg)
    return p


def __DeviceClass__attribute_factory(self, attr_list_wrapper):
    """
    for internal usage only!!!

    Note: attempts to do anything with attr_list_wrapper here are
    not 100% save and may result either in crash or memory leak !!!
    """

    for attr_name, attr_info in self.attr_list.items():
        if isinstance(attr_info, AttrData):
            attr_data = attr_info
        else:
            attr_data = AttrData(attr_name, self.get_name(), attr_info)
        if attr_data.forward:
            self._create_fwd_attribute(
                attr_list_wrapper, attr_data.attr_name, attr_data.att_prop
            )
        else:
            self._create_attribute(
                attr_list_wrapper,
                attr_data.attr_name,
                attr_data.attr_type,
                attr_data.attr_format,
                attr_data.attr_write,
                attr_data.dim_x,
                attr_data.dim_y,
                attr_data.display_level,
                attr_data.polling_period,
                attr_data.memorized,
                attr_data.hw_memorized,
                attr_data.alarm_event_implemented,
                attr_data.alarm_event_detect,
                attr_data.archive_event_implemented,
                attr_data.archive_event_detect,
                attr_data.change_event_implemented,
                attr_data.change_event_detect,
                attr_data.data_ready_event_implemented,
                attr_data.read_method_name,
                attr_data.write_method_name,
                attr_data.is_allowed_name,
                attr_data.att_prop,
            )


def __DeviceClass__command_factory(self):
    """for internal usage only"""
    name = self.get_name()
    class_info = get_class(name)
    deviceimpl_class = class_info[1]

    if not hasattr(deviceimpl_class, "init_device"):
        msg = (
            f"Wrong definition of class {name}\n"
            f"The init_device() method does not exist!"
        )
        Except.throw_exception("PyDs_WrongCommandDefinition", msg, "command_factory()")

    for cmd_name, cmd_info in self.cmd_list.items():
        __create_command(self, deviceimpl_class, cmd_name, cmd_info)


def __create_command(self, deviceimpl_class, cmd_name, cmd_info):
    """for internal usage only"""
    name = self.get_name()

    # check for well defined command info

    # check parameter
    if not isinstance(cmd_info, collections.abc.Sequence):
        msg = (
            f"Wrong data type for value for describing command {cmd_name} in "
            f"class {name}\nMust be a sequence with 2 or 3 elements"
        )
        __throw_create_command_exception(msg)

    if len(cmd_info) < 2 or len(cmd_info) > 3:
        msg = (
            f"Wrong number of argument for describing command {cmd_name} in "
            f"class {name}\nMust be a sequence with 2 or 3 elements"
        )
        __throw_create_command_exception(msg)

    param_info, result_info = cmd_info[0], cmd_info[1]

    if not isinstance(param_info, collections.abc.Sequence):
        msg = (
            f"Wrong data type in command argument for command {cmd_name} in "
            f"class {name}\nCommand parameter (first element) must be a sequence"
        )
        __throw_create_command_exception(msg)

    if len(param_info) < 1 or len(param_info) > 2:
        msg = (
            f"Wrong data type in command argument for command {cmd_name} in "
            f"class {name}\nSequence describing command parameters must contain "
            f"1 or 2 elements"
        )
        __throw_create_command_exception(msg)

    param_type = CmdArgType.DevVoid
    try:
        param_type = CmdArgType(param_info[0])
    except Exception:
        msg = (
            f"Wrong data type in command argument for command {cmd_name} in "
            f"class {name}\nCommand parameter type (first element in first "
            f"sequence) must be a tango.CmdArgType"
        )
        __throw_create_command_exception(msg)

    param_desc = ""
    if len(param_info) > 1:
        param_desc = param_info[1]
        if not is_pure_str(param_desc):
            msg = (
                f"Wrong data type in command parameter for command {cmd_name} in "
                f"class {name}\nCommand parameter description (second element "
                f"in first sequence), when given, must be a string"
            )
            __throw_create_command_exception(msg)

    # Check result
    if not isinstance(result_info, collections.abc.Sequence):
        msg = (
            f"Wrong data type in command result for command {cmd_name} in "
            f"class {name}\nCommand result (second element) must be a sequence"
        )
        __throw_create_command_exception(msg)

    if len(result_info) < 1 or len(result_info) > 2:
        msg = (
            f"Wrong data type in command result for command {cmd_name} in "
            f"class {name}\nSequence describing command result must contain "
            f"1 or 2 elements"
        )
        __throw_create_command_exception(msg)

    result_type = CmdArgType.DevVoid
    try:
        result_type = CmdArgType(result_info[0])
    except Exception:
        msg = (
            f"Wrong data type in command result for command {cmd_name} in "
            f"class {name}\nCommand result type (first element in second "
            f"sequence) must be a tango.CmdArgType"
        )
        __throw_create_command_exception(msg)

    result_desc = ""
    if len(result_info) > 1:
        result_desc = result_info[1]
        if not is_pure_str(result_desc):
            msg = (
                f"Wrong data type in command result for command {cmd_name} in "
                f"class {name}\nCommand parameter description (second element "
                f"in second sequence), when given, must be a string"
            )
            __throw_create_command_exception(msg)

    # If it is defined, get addictional dictionnary used for optional parameters
    display_level, default_command, polling_period, is_allowed = (
        DispLevel.OPERATOR,
        False,
        -1,
        None,
    )

    if len(cmd_info) == 3:
        extra_info = cmd_info[2]
        if not isinstance(extra_info, collections.abc.Mapping):
            msg = (
                f"Wrong data type in command information for command {cmd_name} in "
                f"class {name}\nCommand information (third element in sequence), "
                f"when given, must be a dictionary"
            )
            __throw_create_command_exception(msg)

        if len(extra_info) > 4:
            msg = (
                f"Wrong data type in command information for command {cmd_name} in "
                f"class {name}\nThe optional dictionary can not have more than "
                f"four elements"
            )
            __throw_create_command_exception(msg)

        for info_name, info_value in extra_info.items():
            info_name_lower = info_name.lower()
            if info_name_lower == "display level":
                try:
                    display_level = DispLevel(info_value)
                except Exception:
                    msg = (
                        f"Wrong data type in command information for command {cmd_name} in "
                        f"class {name}\nCommand information for display level is not a "
                        f"tango.DispLevel"
                    )
                    __throw_create_command_exception(msg)
            elif info_name_lower == "default command":
                if not is_pure_str(info_value):
                    msg = (
                        f"Wrong data type in command information for command {cmd_name} in "
                        f"class {name}\nCommand information for default command is not a "
                        f"string"
                    )
                    __throw_create_command_exception(msg)
                v = info_value.lower()
                default_command = v == "true"
            elif info_name_lower == "polling period":
                try:
                    polling_period = int(info_value)
                except Exception:
                    msg = (
                        f"Wrong data type in command information for command {cmd_name} in "
                        f"class {name}\nCommand information for polling period is not an "
                        f"integer"
                    )
                    __throw_create_command_exception(msg)
            elif info_name_lower == "is allowed":
                is_allowed = info_value
                if not is_pure_str(is_allowed):
                    msg = (
                        "Wrong data type in command information for command %s in "
                        "class %s\nCommand information for is allowed function name"
                        "is not an string" % (cmd_name, name)
                    )
                    __throw_create_command_exception(msg)
            elif info_name_lower == "is allowed green_mode":
                pass
            else:
                msg = (
                    f"Wrong data type in command information for command {cmd_name} in "
                    f"class {name}\nCommand information has unknown key "
                    f"{info_name}"
                )
                __throw_create_command_exception(msg)

    # check that the method to be executed exists
    try:
        cmd = getattr(deviceimpl_class, cmd_name)
        if not isinstance(cmd, collections.abc.Callable):
            msg = (
                f"Wrong definition of command {cmd_name} in "
                f"class {name}\nThe object exists in class but is not "
                f"a method!"
            )
            __throw_create_command_exception(msg)
    except AttributeError:
        msg = (
            f"Wrong definition of command {cmd_name} in "
            f"class {name}\nThe command method does not exist!"
        )
        __throw_create_command_exception(msg)

    if is_allowed is None:
        is_allowed_name = f"is_{cmd_name}_allowed"
    else:
        is_allowed_name = is_allowed

    try:
        is_allowed_function = getattr(deviceimpl_class, is_allowed_name)
        if not isinstance(is_allowed_function, collections.abc.Callable):
            msg = (
                f"Wrong definition of command {cmd_name} in "
                f"class {name}\nThe object '{is_allowed_name}' exists in class but is "
                f"not a method!"
            )
            __throw_create_command_exception(msg)
    except Exception:
        is_allowed_name = ""

    self._create_command(
        cmd_name,
        param_type,
        result_type,
        param_desc,
        result_desc,
        display_level,
        default_command,
        polling_period,
        is_allowed_name,
    )


def __DeviceClass__new_device(self, klass, dev_class, dev_name):
    return klass(dev_class, dev_name)


def __DeviceClass__device_factory(self, device_list):
    """for internal usage only"""

    klass = self.__class__
    klass_name = klass.__name__
    info, klass = get_class_by_class(klass), get_constructed_class_by_class(klass)

    if info is None:
        raise RuntimeError(f"Device class '{klass_name}' is not registered")

    if klass is None:
        raise RuntimeError(f"Device class '{klass_name}' as not been constructed")

    deviceClassClass, deviceImplClass, deviceImplName = info
    deviceImplClass._device_class_instance = klass

    tmp_dev_list = []
    for dev_name in device_list:
        device = self._new_device(deviceImplClass, klass, dev_name)
        self._add_device(device)
        tmp_dev_list.append(device)

    self.dyn_attr(tmp_dev_list)

    for dev in tmp_dev_list:
        if Util._UseDb and not Util._FileDb:
            self.export_device(dev)
        else:
            self.export_device(dev, dev.get_name())
    self.py_dev_list += tmp_dev_list


def __DeviceClass__create_device(self, device_name, alias=None, cb=None):
    """
    create_device(self, device_name, alias=None, cb=None) -> None

        Creates a new device of the given class in the database, creates a new
        DeviceImpl for it and calls init_device (just like it is done for
        existing devices when the DS starts up)

        An optional parameter callback is called AFTER the device is
        registered in the database and BEFORE the init_device for the
        newly created device is called

    Throws tango.DevFailed:
        - the device name exists already or
        - the given class is not registered for this DS.
        - the cb is not a callable

    New in PyTango 7.1.2

    Parameters :
        - device_name : (str) the device name
        - alias : (str) optional alias. Default value is None meaning do not create device alias
        - cb : (callable) a callback that is called AFTER the device is registered
               in the database and BEFORE the init_device for the newly created
               device is called. Typically you may want to put device and/or attribute
               properties in the database here. The callback must receive a parameter
               device_name (str). Default value is None meaning no callback

    Return     : None"""
    util = Util.instance()
    util.create_device(self.get_name(), device_name, alias=alias, cb=cb)


def __DeviceClass__delete_device(self, device_name):
    """
    delete_device(self, klass_name, device_name) -> None

        Deletes an existing device from the database and from this running
        server

        Throws tango.DevFailed:
            - the device name doesn't exist in the database
            - the device name doesn't exist in this DS.

    New in PyTango 7.1.2

    Parameters :
        - klass_name : (str) the device class name
        - device_name : (str) the device name

    Return     : None"""
    util = Util.instance()
    util.delete_device(self.get_name(), device_name)


def __DeviceClass__dyn_attr(self, device_list):
    """
    dyn_attr(self,device_list) -> None

        Default implementation does not do anything
        Overwrite in order to provide dynamic attributes

    Parameters :
        - device_list : (seq<DeviceImpl>) sequence of devices of this class

    Return     : None"""
    pass


def __DeviceClass__device_destroyer(self, name):
    """for internal usage only"""
    name = name.lower()
    for d in self.py_dev_list:
        dname = d.get_name().lower()
        if dname == name:
            dev_cl = d.get_device_class()
            # the internal C++ device_destroyer isn't case sensitive so we
            # use the internal DeviceImpl name to make sure the DeviceClass
            # finds it
            dev_cl._device_destroyer(d.get_name())
            self.py_dev_list.remove(d)
            return
    err_mess = "Device " + name + " not in Tango class device list!"
    Except.throw_exception(
        "PyAPI_CantDestroyDevice", err_mess, "DeviceClass.device_destroyer"
    )


def __DeviceClass__device_name_factory(self, dev_name_list):
    """
    device_name_factory(self, dev_name_list) ->  None

        Create device(s) name list (for no database device server).
        This method can be re-defined in DeviceClass sub-class for
        device server started without database. Its rule is to
        initialise class device name. The default method does nothing.

    Parameters :
        - dev_name_list : (seq<str>) sequence of devices to be filled

    Return     : None"""
    pass


def device_class_init():
    DeviceClass.class_property_list = {}
    DeviceClass.device_property_list = {}
    DeviceClass.cmd_list = {}
    DeviceClass.attr_list = {}
    DeviceClass.__init_orig__ = DeviceClass.__init__
    DeviceClass.__init__ = __DeviceClass__init__
    DeviceClass.__str__ = __DeviceClass__str__
    DeviceClass.__repr__ = __DeviceClass__repr__
    DeviceClass._create_user_default_attr_prop = (
        __DeviceClass__create_user_default_attr_prop
    )
    DeviceClass._attribute_factory = __DeviceClass__attribute_factory
    DeviceClass._command_factory = __DeviceClass__command_factory
    DeviceClass._new_device = __DeviceClass__new_device

    DeviceClass.device_factory = __DeviceClass__device_factory
    DeviceClass.create_device = __DeviceClass__create_device
    DeviceClass.delete_device = __DeviceClass__delete_device
    DeviceClass.dyn_attr = __DeviceClass__dyn_attr
    DeviceClass.device_destroyer = __DeviceClass__device_destroyer
    DeviceClass.device_name_factory = __DeviceClass__device_name_factory



#!/usr/bin/env python3
# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

"""
Trivial power supply device with no external connection or behaviour.

Extends ps0b.py with a DevEnum attribute "output_tracking".
This would make sense in dual-output power supply, but that
hasn't been implemented here, for simplicity.
"""

import enum
import random
from time import sleep
from tango import AttrWriteType, DebugIt, InfoIt
from tango.server import Device, attribute, command


class TrackingMode(enum.IntEnum):
    INDEPENDENT = 0  # must start at zero!
    SYNCED = 1  # and increment by 1


class PowerSupply(Device):
    _tracking_mode = TrackingMode.SYNCED

    @attribute(dtype=TrackingMode, access=AttrWriteType.READ_WRITE)
    def output_tracking(self):
        return self._tracking_mode

    @output_tracking.write
    def output_tracking(self, value):
        self._tracking_mode = value

    @attribute(dtype=float, polling_period=3000, rel_change=1e-3)  # milliseconds
    @DebugIt(show_args=False, show_kwargs=False, show_ret=True)
    def voltage(self):
        noise = -0.05 + 0.1 * random.random()
        return 1.5 + noise

    @command
    @InfoIt()
    def calibrate(self):
        sleep(0.1)


if __name__ == "__main__":
    PowerSupply.run_server()



#!/usr/bin/env python
# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

"""
Simple client to show how to connect to a Clock device from ClockDS

usage: client clock_dev_name
"""

import sys
import tango

if len(sys.argv) != 2:
    print("must provide one and only one clock device name")
    sys.exit(1)

clock = tango.DeviceProxy(sys.argv[1])
t = clock.time
gmt = clock.gmtime
noon = clock.noon
display = clock.display
print(t)
print(gmt)
print(noon, noon.name, noon.value)
if noon == noon.AM:
    print("Good morning!")
print(clock.ctime(t))
print(clock.mktime(gmt))
print(display, display.name, display.value)
clock.display = display.ANALOG
clock.display = "DIGITAL"  # you can use a valid string to set the value
print(clock.display, clock.display.name, clock.display.value)
display_type = type(display)  # or even create your own IntEnum type
analog = display_type(0)
clock.display = analog
print(clock.display, clock.display.name, clock.display.value)
clock.display = clock.display.DIGITAL
print(clock.display, clock.display.name, clock.display.value)



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
"""
Basic unit test for a PowerSupply device with events.  Requires pytest.
"""

import time
from tango import EventType
from tango.test_utils import DeviceTestContext
from pytest import approx

from ps0b import PowerSupply


def test_calibrate():
    """Test device calibration and voltage reading."""
    with DeviceTestContext(PowerSupply, process=True) as proxy:
        proxy.calibrate()
        assert proxy.voltage == approx(1.5, abs=0.1)


def test_events():
    """Test change events occur."""
    results = []

    def callback(evt):
        if not evt.err:
            results.append(evt)

    with DeviceTestContext(PowerSupply, process=True) as proxy:
        eid = proxy.subscribe_event(
            "voltage", EventType.CHANGE_EVENT, callback, wait=True
        )
        # wait for events to happen
        time.sleep(4)  # not ideal in tests!
        assert len(results) > 1
        proxy.unsubscribe_event(eid)
